@@ -451,7 +451,7 @@ _internal_convert_v0 (rtcom_el_db_t db)
 
   /* Now, if we've updated S3 db, we have to clean up some old triggers; this
    * is copy/paste from the db initialisation. */
-  
+
   runsql ("DROP TRIGGER IF EXISTS gc_update_ev_add2;");
   runsql ("DROP TRIGGER IF EXISTS gc_update_ev_add3;");
   runsql ("DROP TRIGGER IF EXISTS gc_update_ev_update;");
@@ -500,64 +500,76 @@ rtcom_el_db_convert_from_db0 (const gchar *fname, const gchar *old_fname)
   gint ret;
   int cnt;
   gboolean success;
+  gchar *temp_fname;
+  GError *err = NULL;
 
   /* To speed things up in most likely case (new database already exists),
    * we first check whether the new database file already exists. */
   if (g_access (fname, 0) == 0)
       return TRUE;
 
-  /* We don't want to start converting into a new database if we can't
-   * reach the old one. */
+  /* If we can't reach the old database, there's nothing for us to do. */
   if (g_access (old_fname, 0) == -1)
-      return FALSE;
+      return TRUE;
 
   ret = sqlite3_open (old_fname, &old_db);
 
+  /* Can't even open old db (no way of telling whether it's corrupted or
+   * not as of yet), so just ignore it. */
   if (ret != SQLITE_OK)
-      return FALSE;
-
-  ret = sqlite3_open (fname, &db);
-  if (ret != SQLITE_OK)
-    {
-      sqlite3_close (old_db);
-      return FALSE;
-    }
-
-  /* We'd rather be a bit slower with on-disk journal then screw up db
-   * containing user data. */
-  rtcom_el_db_exec (db, NULL, NULL, "PRAGMA journal_mode = DELETE;", NULL);
-
-  /* Let's reduce caches size from 8MB (default 2000 pages) to 0.5MB
-   * (128 pages);
-   */
-  rtcom_el_db_exec (db, NULL, NULL, "PRAGMA cache_size = 128;", NULL);
-
-  if (!rtcom_el_db_exec (db, rtcom_el_db_single_int, &ret,
-      "PRAGMA user_version", NULL))
-    {
-      sqlite3_close (old_db);
-      sqlite3_close (db);
-      return FALSE;
-    }
-
-  /* It's already been converted. */
-  if (ret == 1)
-    {
-      sqlite3_close (old_db);
-      sqlite3_close (db);
       return TRUE;
+
+  temp_fname = g_strconcat (fname, ".temp", NULL);
+
+  ret = sqlite3_open (temp_fname, &db);
+  if (ret != SQLITE_OK)
+    {
+      sqlite3_close (old_db);
+      g_free (temp_fname);
+      return FALSE;
     }
 
-  /* While we're busy converting the database, another app might notice
-   * the file exists and try to open it. If user_version is not set
-   * immediately, it would fail miserably. */
-  rtcom_el_db_exec (db, NULL, NULL, "PRAGMA user_version = 1;", NULL);
+  /* We're doing the conversion to a temporary db, so it doesn't
+   * matter if we get interrupted. But we want to be done as quickly
+   * as possible to avoid it (and make a successfull conversion
+   * in the process).
+   * These settings are active only during the conversion, since the
+   * database gets closed and reopened afterwards. */
+  rtcom_el_db_exec (db, NULL, NULL, "PRAGMA journal_mode = OFF;", NULL);
 
-  if (!rtcom_el_db_transaction (db, TRUE, NULL))
+  /* We try to get an exclusive lock on the new database. If we fail,
+   * due to db/table being locked, this means another process is
+   * doing the conversion. */
+  if (!rtcom_el_db_transaction (db, TRUE, &err))
     {
       sqlite3_close (old_db);
       sqlite3_close (db);
-      return FALSE;
+
+      /* If the failure is due to db format error, this means another
+       * process was killed while database was being created, we may
+       * safely overwrite it. */
+      if ((err->code == SQLITE_CORRUPT) ||
+          (err->code == SQLITE_FORMAT) ||
+          (err->code == SQLITE_NOTADB) ||
+          (err->code == SQLITE_ABORT) ||
+          (err->code == SQLITE_INTERRUPT))
+        {
+          g_warning ("%s: temporary db corrupted, redoing upgrade", G_STRFUNC);
+          g_unlink (temp_fname);
+          g_free (temp_fname);
+          g_error_free (err);
+
+          /* with corrupted temp file out of the way, retry */
+          return rtcom_el_db_convert_from_db0 (fname, old_fname);
+        }
+      else
+        {
+          /* we better wait until some other process converts it */
+          g_warning ("%s: database upgrade in progress, will wait", G_STRFUNC);
+          g_free (temp_fname);
+          g_error_free (err);
+          return FALSE;
+        }
     }
 
   /* Start backing up old database into a new one (with the same format). */
@@ -566,6 +578,8 @@ rtcom_el_db_convert_from_db0 (const gchar *fname, const gchar *old_fname)
     {
       sqlite3_close (old_db);
       sqlite3_close (db);
+      g_unlink (temp_fname);
+      g_free (temp_fname);
       return FALSE;
     }
 
@@ -583,10 +597,13 @@ rtcom_el_db_convert_from_db0 (const gchar *fname, const gchar *old_fname)
 
   ret = sqlite3_backup_finish (bkp);
 
+  /* We couldn't make it. Better remove the corrupted file, too. */
   if (ret != SQLITE_OK)
     {
       sqlite3_close (old_db);
       sqlite3_close (db);
+      g_unlink (temp_fname);
+      g_free (temp_fname);
       return FALSE;
     }
 
@@ -596,6 +613,20 @@ rtcom_el_db_convert_from_db0 (const gchar *fname, const gchar *old_fname)
   success = _internal_convert_v0 (db);
 
   sqlite3_close (db);
+
+  /* If everything went OK, the database was successfully converted.
+   * Let's move it to correct filename. Otherwise, remove the
+   * corrupted conversion file and give up. */
+  if (success)
+    {
+      g_rename (temp_fname, fname);
+    }
+  else
+    {
+      g_unlink (temp_fname);
+    }
+
+  g_free (temp_fname);
   return success;
 }
 
