@@ -227,8 +227,15 @@ static gboolean _ensure_db(RTComEl *el, gboolean reopen)
         OLD_SQLITE_DATABASE, NULL);
 
     /* If needed, this will convert the v0 database format into the
-     * one we use today. */
-    rtcom_el_db_convert_from_db0 (fn, old_fn);
+     * one we use today. If this fails for any reason, we shouldn't
+     * just create a new empty db, instead we should wait for some
+     * other process (or perhaps us later) to do the upgrade. */
+    if (!rtcom_el_db_convert_from_db0 (fn, old_fn))
+    {
+        g_free (old_fn);
+        g_free (fn);
+        return FALSE;
+    }
 
     g_free (old_fn);
 
